<?php

namespace Drupal\prometheus_metrics\EventSubscriber;

use Psr\Log\LoggerInterface;
use Drupal\Component\Utility\Timer;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\prometheus_metrics\Bridge\PrometheusMetricsInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Event\GetResponseEvent;
use Symfony\Component\HttpKernel\Event\PostResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Counts all requests/ total requests and provides request timer metrics
 * Class PrometheusRequestSubscriber
 * @package Drupal\prometheus_metrics\EventSubscriber
 */
class PrometheusRequestSubscriber implements EventSubscriberInterface
{

  /**
   * @var RouteMatchInterface
   */
  private $routeMatch;

  /**
   * @var PrometheusMetricsInterface
   */
  private $prometheusMetrics;

  /**
   * @var ImmutableConfig
   */
  private $config;

  /**
   * @var string
   */
  private $namespace;
  /**
   * @var LoggerInterface
   */
  private $logger;

  /**
   * PrometheusRequestSubscriber constructor.
   * @param PrometheusMetricsInterface $prometheusMetrics
   * @param RouteMatchInterface $routeMatch
   * @param ConfigFactoryInterface $configFactory
   * @param LoggerInterface $logger
   */
  public function __construct(
    PrometheusMetricsInterface $prometheusMetrics,
    RouteMatchInterface $routeMatch,
    ConfigFactoryInterface $configFactory,
    LoggerInterface $logger
  ) {
    $this->routeMatch = $routeMatch;
    $this->prometheusMetrics = $prometheusMetrics;
    $this->config = $configFactory->get(PrometheusDefaults::CONFIGURATION_NAME);
    $this->namespace = $this->config->get('metrics_namespace')?$this->config->get('metrics_namespace'):PrometheusDefaults::METRICS_NAMESPACE;
    $this->logger = $logger;
  }

  public static function getSubscribedEvents()
  {
    return [
      KernelEvents::REQUEST => ['onRequest',1000],
      KernelEvents::TERMINATE => ['onTerminate',0]
    ];
  }

  /**
   * Start timer for the request
   * @param GetResponseEvent $event
   */
  public function onRequest(GetResponseEvent $event) {
    Timer::start(PrometheusDefaults::TIMER_NAME);
  }

  /**
   * Handles the terminate event, storing a histogram timer and
   * counter for the request by method, route and status.
   *
   * @param PostResponseEvent $event
   */
  public function onTerminate(PostResponseEvent $event){
    Timer::stop(PrometheusDefaults::TIMER_NAME);
    $timeInMs = Timer::read(PrometheusDefaults::TIMER_NAME);
    $timeInS = $timeInMs/1000;

    $method = $event->getRequest()->getMethod();
    $routeName = $this->getRouteName($event->getRequest());

    //calls function to get/register counter based on namespace and name
    $counter = $this->prometheusMetrics->getCounter(
      $this->namespace,
      'http_requests_total',
      'Total number of requests',
      [
        'method',
        'route',
        'status'
      ]
    );

    $counter->inc([$method, $routeName, $this->flattenRepsonseCode($event->getResponse())]);

    //calls function to get/register histogram based on namespace and name
    $histogram = $this->prometheusMetrics->getHistogram(
      $this->namespace,
      'http_requests_total',
      'Timer for all requests',
      [
        'method',
        'route',
        'status'
      ]
    );

    $histogram->observe($timeInS,[$method, $routeName, $this->flattenRepsonseCode($event->getResponse())]);
  }

  /**
   * Extracts the route name from the request
   * or logs a warning  if the route is empty
   *
   * @param Request $request
   * @return string|string[]|null
   */
  private function getRouteName(Request $request){
    //does routematch return a value
    if(!empty($this->routeMatch->getRouteName())){
      return str_replace(".","_",$this->routeMatch->getRouteName());
    }else{
      //@todo understand which requests fall into this and provide better categorisation for metrics
      $this->logger->warning('Route name not found',
        [
          'requestUri'=>$request->getRequestUri(),
          'method'=>$request->getMethod()
        ]
      );
      return '';
    }
  }

  /**
   * Flattens Response Status codes into 4 strings 2xx - 5xx
   *
   * @param Response $response
   * @return int|string
   */
  private function flattenRepsonseCode(Response $response){
    if ($response->getStatusCode() >= 200 && $response->getStatusCode() < 300) {
      return '2xx';
    } elseif ($response->getStatusCode() >= 300 && $response->getStatusCode() < 400) {
      return '3xx';
    } elseif ($response->getStatusCode() >= 400 && $response->getStatusCode() < 500) {
      return '4xx';
    } elseif ($response->getStatusCode() >= 500) {
      return '5xx';
    }
    return $response->getStatusCode();
  }
}
